@@ -6,6 +6,9 @@
 \i plr.sql
 \set ECHO all
 
+-- make sure we get the notices
+set client_min_messages to notice;
+
 -- check version
 SELECT plr_version();
 
@@ -294,7 +297,7 @@ if (arg2 < 1 || arg3 < 1 || arg4 < 1)
 if (arg2 > dim(arg1)[1] || arg3 > dim(arg1)[2] || arg4 > dim(arg1)[3])
   return(NA)
 return(arg1[arg2,arg3,arg4])
-' language 'plr' WITH (isstrict);
+' language 'plr' STRICT;
 
 select arr3d('{{{111,112},{121,122},{131,132}},{{211,212},{221,222},{231,232}}}',2,3,1) as "231";
 -- for sake of comparison, see what normal pgsql array operations produces
@@ -309,7 +312,7 @@ select f1[0][1][1] is null as "NULL" from (select '{{{111,112},{121,122},{131,13
 --
 -- test 3D array return value
 --
-create or replace function arr3d(_int4) returns int4[] as 'return(arg1)' language 'plr' WITH (isstrict);
+create or replace function arr3d(_int4) returns int4[] as 'return(arg1)' language 'plr' STRICT;
 select arr3d('{{{111,112},{121,122},{131,132}},{{211,212},{221,222},{231,232}}}');
 
 --
@@ -430,7 +433,7 @@ returns setof record as '
  return (arg1)
 ' language 'plr';
 
-select * from restore_df((select test_serialize('select oid, typname from pg_type where typname in (''oid'',''name'',''int4'')'))) as t(oid oid, typname name);
+select * from restore_df((select test_serialize('select oid, typname from pg_type where typname in (''oid'',''name'',''int4'')'))) as t(oid oid, typname name) order by oid;
 
 --Test WINDOW functions
 -- create test table
@@ -453,7 +456,7 @@ RETURNS float8 AS
 $BODY$
   slope <- NA
   y <- farg1
-  x <- farg2 
+  x <- farg2
   if (fnumrows==9) try (slope <- lm(y ~ x)$coefficients[2])
   return(slope)
 $BODY$
@@ -465,3 +468,44 @@ FROM (SELECT firm, fyear, eps,
 FROM test_data) AS a
 WHERE eps IS NOT NULL
 WINDOW w AS (ORDER BY firm, fyear ROWS 8 PRECEDING);
+
+CREATE OR REPLACE FUNCTION rlargeint8out(n int) RETURNS int8[] AS $$
+matrix(2, 1, n)
+$$ LANGUAGE plr;
+
+CREATE OR REPLACE FUNCTION routfloat4(n int) RETURNS float4[] AS $$
+vector(mode = "numeric", length = n)
+$$ LANGUAGE plr;
+
+SELECT rlargeint8out(10);
+SELECT routfloat4(10);
+
+SELECT count(rlargeint8out(15000));
+SELECT count(routfloat4(15000));
+
+CREATE table tbl(val integer);
+
+CREATE OR REPLACE FUNCTION test_create_procedure() RETURNS void
+  AS
+$BODY$
+  version_11plus  <- pg.spi.exec("select current_setting('server_version_num')::integer >= 110000;")
+  if(version_11plus[[1]]) {
+    pg.spi.exec("
+      CREATE OR REPLACE PROCEDURE insert_data(a int, b int)
+        AS
+      $$
+        pg.spi.exec('INSERT INTO tbl VALUES (1);')
+        pg.spi.exec('INSERT INTO tbl VALUES (2);')
+      $$ LANGUAGE plr;
+      ")
+    pg.spi.exec("CALL insert_data(1, 2);")
+  } else {
+    pg.spi.exec("INSERT INTO tbl VALUES (1);")
+    pg.spi.exec("INSERT INTO tbl VALUES (2);")
+  }
+$BODY$
+LANGUAGE plr;
+
+SELECT test_create_procedure();
+
+SELECT * FROM tbl;
